SystemOrganization addCategory: #'CS598rej-Payroll'!
SystemOrganization addCategory: #'CS598rej-Payroll-Test'!

!SortedCollectionTest methodsFor: '*CS598rej-Payroll-Test' stamp: 'm3r 6/6/2008 22:15'!
testIndexOfStartOfIntervalContaining
	| sequence |
	sequence := #(1 80 500 501) asSortedCollection.
	
	self assert: (sequence indexOfStartOfIntervalContaining: 1) = 1.
	self assert: (sequence indexOfStartOfIntervalContaining: 79) = 1.
	self assert: (sequence indexOfStartOfIntervalContaining: 80) = 2.
	self assert: (sequence indexOfStartOfIntervalContaining: 501)	= 4
! !

!SortedCollection methodsFor: '*CS598rej-Payroll' stamp: ''!
indexOfStartOfIntervalContaining: anElement 
	"Return the index of anElement or, if it is not present, of the index of the
	largest element smaller than it."

	self isEmpty ifTrue: [^0].
	^(self indexForInserting: anElement) - firstIndex
! !

TestCase subclass: #EmployeeTest
	instanceVariableNames: 'susan day1 normalMonth aLittleOvertime'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'CS598rej-Payroll-Test'!

!EmployeeTest commentStamp: 'm3r 6/6/2008 22:19' prior: 0!
An EmployeeTest is xxxxxxxxx.

Instance Variables
	aLittleOvertime:		<Object>
	day1:				<Object>
	normalMonth:		<Object>
	susan:				<Object>

aLittleOvertime
	- xxxxx

day1
	- xxxxx

normalMonth
	- xxxxx

susan
	- xxxxx
!

!EmployeeTest methodsFor: 'Running' stamp: 'dd 1/11/2005 19:36'!
setUp
	day1 := Date 
				newDay: 11
				month: 7
				year: 2001.
	susan := Employee new named: 'Susan'.
	normalMonth := #(#(40 0) #(40 0) #(40 0) #(40 0)).
	aLittleOvertime := #(#(40 0) #(50 0) #(40 0) #(40 0))
! !

!EmployeeTest methodsFor: 'Testing' stamp: 'dd 12/23/2004 19:45'!
testChangeSalary
	susan changeSalaryFor: day1 to: 20.
	self assert: susan salary = 20.
	day1 addDays: 7.
	susan changeSalaryFor: day1 to: 25.
	self assert: susan salary = 25
! !

TestCase subclass: #PayrollSystemTest
	instanceVariableNames: 'myPayroll day1 susan george normalMonth aLittleOvertime'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'CS598rej-Payroll-Test'!

!PayrollSystemTest commentStamp: 'm3r 6/6/2008 22:20' prior: 0!
A PayrollSystemTest is xxxxxxxxx.

Instance Variables
	aLittleOvertime:		<Object>
	day1:				<Object>
	george:				<Object>
	myPayroll:			<Object>
	normalMonth:		<Object>
	susan:				<Object>

aLittleOvertime
	- xxxxx

day1
	- xxxxx

george
	- xxxxx

myPayroll
	- xxxxx

normalMonth
	- xxxxx

susan
	- xxxxx
!

!PayrollSystemTest methodsFor: 'private' stamp: ''!
postTimeCards: collection to: anEmployee 
	| timecardDay |
	timecardDay := day1.
	collection do: 
			[:each | 
			anEmployee 
				postTimeCardFor: timecardDay
				hoursWorked: each first
				vacation: each last.
			timecardDay := timecardDay addDays: 7]
! !

!PayrollSystemTest methodsFor: 'Running' stamp: 'dd 1/11/2005 19:38'!
setUp
	myPayroll := PayrollSystem new name: 'ACME'.
	day1 := Date newDay: 11 month: 7 year: 2001. 
	susan := Employee new named: 'Susan'.
	susan changeSalaryFor: day1 to: 20.
	george := Employee new named: 'George'.
	george changeSalaryFor: day1 to: 12.
	normalMonth := #(#(40 0) #(40 0) #(40 0) #(40 0)).
	aLittleOvertime := #(#(40 0) #(50 0) #(40 0) #(40 0))
! !

!PayrollSystemTest methodsFor: 'Testing' stamp: ''!
testAddEmployee
	myPayroll addEmployee: susan.
	self assert: (myPayroll employeeNames includes: 'Susan').
	myPayroll addEmployee: george.
	self assert: (myPayroll employeeNamed: 'Susan') = susan
! !

!PayrollSystemTest methodsFor: 'Testing' stamp: ''!
testEarnings
	myPayroll
		addEmployee: susan;
		addEmployee: george.
	self postTimeCards: normalMonth to: susan.
	self postTimeCards: aLittleOvertime to: george.
	self assert: susan earned = 3200.
	self assert: george earned = 2100
! !

!PayrollSystemTest methodsFor: 'testing' stamp: 'dd 1/20/2005 22:14'!
testTakeHomePay
	| firstDay nextDay paycheckDay john hireDate paycheck |
	
	john := Employee named: 'John Doe'.
	hireDate := Date newDay: 1 month: 5 year: 1999.
	john changeSalaryFor: hireDate to: 20.
	
	firstDay := Date newDay: 5 month: 5 year: 1999.
	john postTimeCardFor: firstDay hoursWorked: 40 vacation: 0.
	nextDay := Date newDay: 12 month: 5 year: 1999.
	john postTimeCardFor: nextDay hoursWorked: 40 vacation: 0.
	
	paycheckDay := Date newDay: 15 month: 5 year: 1999.
	paycheck := john makePaycheckAt: paycheckDay.
	
	self assert: paycheck amountPaid = 1188.8
! !

Object subclass: #Employee
	instanceVariableNames: 'name transactions salary earned paid withholding accruedVacation taxRule'
	classVariableNames: 'HeadTaxRate MarriedSeparateTaxRate MarriedTaxRate SingleTaxRate'
	poolDictionaries: ''
	category: 'CS598rej-Payroll'!

!Employee commentStamp: 'm3r 6/6/2008 22:19' prior: 0!
I represent an employee in a payroll system.  Thus, I keep track of the number of hours worked,
the wages paid, the vacation accrued, the taxes withheld, and so on.  My values change only
when transactions are posted to me.  Transactions are subclasses of EmployeeTransaction.

Variables
	name 				<String>
	transactions 		<Collection of Transactions, sorted by date>
	salary 				<ValueWithHistory>
	earned  			<ValueWithHistory>
	paid  				<ValueWithHistory>
	withholding  		<ValueWithHistory>
	accruedVacation  	<ValueWithHistory>
	taxRule 			<TaxRule>!

!Employee class methodsFor: 'class initialization' stamp: ''!
initialize
	"Employee initialize"
	SingleTaxRate := TaxRule new.
	SingleTaxRate
		breakPoint: 0 rate: 0.15;
		breakPoint: 22750 rate: 0.28;
		breakPoint: 55100 rate: 0.31;
		breakPoint: 115000 rate: 0.36;
		breakPoint: 250000 rate: 0.396.

	MarriedTaxRate := TaxRule new.
	MarriedTaxRate
		breakPoint: 0 rate: 0.15;
		breakPoint: 38000 rate: 0.28;
		breakPoint: 91850 rate: 0.31;
		breakPoint: 140000 rate: 0.36;
		breakPoint: 250000 rate: 0.396.

	MarriedSeparateTaxRate := TaxRule new.
	MarriedSeparateTaxRate
		breakPoint: 0 rate: 0.15;
		breakPoint: 19500 rate: 0.28;
		breakPoint: 45925 rate: 0.31;
		breakPoint: 70000 rate: 0.36;
		breakPoint: 125000 rate: 0.396.

	HeadTaxRate := TaxRule new.
	HeadTaxRate
		breakPoint: 0 rate: 0.15;
		breakPoint: 30500 rate: 0.28;
		breakPoint: 78700 rate: 0.31;
		breakPoint: 127500 rate: 0.36;
		breakPoint: 250000 rate: 0.396.
! !

!Employee class methodsFor: 'instance creation' stamp: 'm3r 6/20/2008 17:14'!
named: nameString
	^ self new named: nameString
! !

!Employee methodsFor: 'actions' stamp: ''!
changeSalaryFor: aDate to: dollarPerHour 
	self postTransaction: (SalaryChange new
			date: aDate
			employee: self
			salary: dollarPerHour)
! !

!Employee methodsFor: 'accessing' stamp: ''!
earned
	^earned
! !

!Employee methodsFor: 'private' stamp: ''!
federalTaxesOn: money assumingTotalIs: estimatedEarnings
	^money * (taxRule rateFor: estimatedEarnings)
! !

!Employee methodsFor: 'posting' stamp: ''!
incrementEarned: anAmount
	earned := earned +  anAmount
! !

!Employee methodsFor: 'posting' stamp: ''!
incrementPaid: anAmount
	paid := paid + anAmount
! !

!Employee methodsFor: 'posting' stamp: ''!
incrementWithholding: anAmount
	withholding := withholding + anAmount
! !

!Employee methodsFor: 'actions' stamp: ''!
makePaycheckAt: aDate 
	"Create a paycheck for the amount owed to the employee. It hasn't 
	been posted."

	| amountPaid amountWithheld |
	amountPaid := earned - paid - withholding.
	amountWithheld := self taxesOn: amountPaid at: aDate.
	amountPaid := amountPaid - amountWithheld.
	^Paycheck new
		date: aDate
		employee: self
		paid: amountPaid
		withheld: amountWithheld
		totalPaid: paid + amountPaid
		totalWithheld: withholding + amountWithheld
! !

!Employee methodsFor: 'accessing' stamp: ''!
name
	^name
! !

!Employee methodsFor: 'initialize-release' stamp: ''!
named: aString
	name := aString.
	transactions := OrderedCollection new.
	salary := 0.
	earned := 0.
	paid := 0.
	withholding := 0.
	accruedVacation := 0.
	taxRule := SingleTaxRate
! !

!Employee methodsFor: 'accessing' stamp: ''!
paid
	^paid
! !

!Employee methodsFor: 'actions' stamp: ''!
postTimeCardFor: aDate hoursWorked: hoursW vacation: hoursV 
	self postTransaction: (Timecard new
			date: aDate
			employee: self
			worked: hoursW
			vacation: hoursV)
! !

!Employee methodsFor: 'posting' stamp: ''!
postTransaction: aTransaction
	"Public interface to record and process (i.e. post) a transaction."
	aTransaction postTo: self.
	transactions add: aTransaction
! !

!Employee methodsFor: 'accessing' stamp: ''!
salary
	^salary
! !

!Employee methodsFor: 'accessing' stamp: ''!
salary: anAmount
	salary := anAmount
! !

!Employee methodsFor: 'private' stamp: ''!
taxesOn: money at: date 
	"Return the taxes to be withheld from the money earned."
	"This is just a guess - I didn't look up the tax rules."

	| socialSecurity federalTaxes stateTaxes estimatedEarnings |
	socialSecurity := (0 max: (money min: 60600 - earned))
				* 0.077.
	estimatedEarnings := earned + money * date day / date daysInYear.
	federalTaxes := self federalTaxesOn: money assumingTotalIs: estimatedEarnings.
	stateTaxes := money * 0.03.	"Illinois has a flat tax."
	^socialSecurity + federalTaxes + stateTaxes
! !

Object subclass: #EmployeeTransaction
	instanceVariableNames: 'date employee'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'CS598rej-Payroll'!

!EmployeeTransaction commentStamp: 'm3r 6/6/2008 22:19' prior: 0!
I'm an abstract class of transactions that can be posted to an Employee.

My most important method is postTo:, which is implemented by subclasses
to send a message to the Employee to post the transaction.  The Employee
has all the information, so it has to be the one to implement posting, but the
postTo: method implements a standard interface so that clients of the
payroll system can always post a transaction by sending the postTransaction:
message to an Employee.

Variables
	date			<Date>   		Date at which transaction occured.
	employee	<Employee>

!

!EmployeeTransaction methodsFor: 'accessing' stamp: ''!
date
	^date
! !

!EmployeeTransaction methodsFor: 'posting' stamp: ''!
post
	employee postTransaction: self
! !

!EmployeeTransaction methodsFor: 'posting' stamp: ''!
postTo: anEmployee
	"double dispatching"
	self subclassResponsibility
! !

EmployeeTransaction subclass: #Paycheck
	instanceVariableNames: 'amountPaid taxes totalPaid totalTaxes'
	classVariableNames: 'AmountFormat DateFormat'
	poolDictionaries: ''
	category: 'CS598rej-Payroll'!

!Paycheck commentStamp: '<historical>' prior: 0!
Transaction representing a payment to an employee.!

!Paycheck class methodsFor: 'instance creation' stamp: 'm3r 6/20/2008 17:17'!
date: aDate employee: anEmployee paid: amount1 withheld: amount2 totalPaid: amount3 totalWithheld: amount4
	^ self new 
		date: aDate 
		employee: anEmployee 
		paid: amount1
		withheld: amount2 
		totalPaid: amount3 
		totalWithheld: amount4
! !

!Paycheck methodsFor: 'accessing' stamp: ''!
amountPaid
	^amountPaid
! !

!Paycheck methodsFor: 'initialize-release' stamp: ''!
date: aDate employee: anEmployee paid: amount1 withheld: amount2 totalPaid: amount3 totalWithheld: amount4
	date := aDate.
	employee := anEmployee.
	amountPaid := amount1.
	taxes := amount2.
	totalPaid := amount3.
	totalTaxes := amount4
! !

!Paycheck methodsFor: 'posting' stamp: ''!
postTo: anEmployee
	anEmployee incrementPaid: amountPaid.
	anEmployee incrementWithholding: taxes
! !

!Paycheck methodsFor: 'printing' stamp: 'dd 1/24/2005 15:49'!
printOnCheckStream: aStream
	"This method is supposed to print information on a check that is on the printer, and
	aStream is the printer.  Unfortunately, the whole method is bogus, because we don't
	know what the printer or the check are like.  I'm assuming that the check has two-parts.
	One is the actual check, and we need to print the employee's name and the amount.
	The second is the check stub, which will say how much is being paid and how much
 	is being withheld. I'm also assuming that the check is preprinted with the name of the
	company, and everything else that is usual for a check."

	aStream cr; cr.
	aStream next: 40 put: (Character space).
	date printOn: aStream.
	aStream cr.
	aStream nextPutAll: ((String new: 40 withAll: Character space) copyReplaceFrom: 1 to: employee name size with: employee name).
	amountPaid printOn: aStream.
	aStream cr; cr; cr; cr.
	aStream nextPutAll: 'this month  '.
	aStream nextPutAll: ' amountPaid: $'.
	amountPaid printOn: aStream.
	aStream nextPutAll: ' taxes: $'.
	taxes printOn: aStream.
	aStream cr.
	aStream nextPutAll: 'year to date '.
	aStream nextPutAll: ' amountPaid: $'.
	totalPaid printOn: aStream.
	aStream nextPutAll: ' taxes: $'.
	totalTaxes printOn: aStream.
	aStream cr; cr.
! !

!Paycheck methodsFor: 'accessing' stamp: 'dd 1/24/2005 15:49'!
taxes
	^taxes 
! !

EmployeeTransaction subclass: #SalaryChange
	instanceVariableNames: 'newSalary'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'CS598rej-Payroll'!

!SalaryChange commentStamp: '<historical>' prior: 0!
Transaction representing a change in salary, usually a raise.!

!SalaryChange methodsFor: 'initialize-release' stamp: ''!
date: aDate employee: anEmployee salary: dollarPerHour
	date := aDate.
	employee := anEmployee.
	newSalary := dollarPerHour
! !

!SalaryChange methodsFor: 'accessing' stamp: ''!
newSalary
	^newSalary
! !

!SalaryChange methodsFor: 'posting' stamp: ''!
postTo: anEmployee
	anEmployee salary: newSalary
! !

EmployeeTransaction subclass: #Timecard
	instanceVariableNames: 'hoursWorked hoursVacation'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'CS598rej-Payroll'!

!Timecard commentStamp: '<historical>' prior: 0!
Transaction representing the work that the employee has done.!

!Timecard class methodsFor: 'instance creation' stamp: 'm3r 6/20/2008 17:25'!
date: aDate employee: anEmployee worked: workHours vacation: vacationHours
	^ self new 
		date: aDate 
		employee: anEmployee 
		worked: workHours
		vacation: vacationHours
! !

!Timecard methodsFor: 'initialize-release' stamp: ''!
date: aDate employee: anEmployee worked: hoursW vacation: hoursV
	date := aDate.
	employee := anEmployee.
	hoursWorked := hoursW.
	hoursVacation := hoursV
! !

!Timecard methodsFor: 'accessing' stamp: ''!
hoursVacation
	^hoursVacation
! !

!Timecard methodsFor: 'accessing' stamp: ''!
hoursWorked
	^hoursWorked
! !

!Timecard methodsFor: 'posting' stamp: ''!
postTo: anEmployee 
	| money overtime nonovertime |
	overtime := hoursWorked - 40 max: 0.
	nonovertime := hoursWorked min: 40.
	money := overtime * 1.5 + nonovertime * anEmployee salary.
	anEmployee incrementEarned: money
! !

Object subclass: #PayrollSystem
	instanceVariableNames: 'employees paychecks companyName'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'CS598rej-Payroll'!

!PayrollSystem commentStamp: '<historical>' prior: 0!
I am the top-level of the payroll system.  I keep track of the employees
and can "run the payroll", i.e. write paychecks for the employees.  To
run the payroll, send the messages
	makePaychecksAt: paydate	- to create the paycheck transactions
	printPaychecksOn: aStream	- to print them out 
	postPaychecks				- to record them in employees

Note that the employees are not marked with having been paid until
the paychecks are posted, so just creating the transactions or even
printing them is not enough.

Variables
	employees   <Dictionary from String to Employee>
	paychecks 	<Collection of Paycheck>	holds paychecks that have been created until they have been posted!

!PayrollSystem class methodsFor: 'instance creation' stamp: ''!
new
	^super new initialize
! !

!PayrollSystem methodsFor: 'accessing' stamp: ''!
addEmployee: anEmployee
	employees at: anEmployee name put: anEmployee
! !

!PayrollSystem methodsFor: 'accessing' stamp: ''!
employeeNamed: aName
	^employees at: aName
! !

!PayrollSystem methodsFor: 'accessing' stamp: ''!
employeeNames
	^employees keys asOrderedCollection
! !

!PayrollSystem methodsFor: 'initialize-release' stamp: ''!
initialize
	employees := Dictionary new.
! !

!PayrollSystem methodsFor: 'actions' stamp: ''!
makePaychecksAt: aDate 
	paychecks := employees collect: [:eachEmployee | eachEmployee makePaycheckAt: aDate]
! !

!PayrollSystem methodsFor: 'accessing' stamp: ''!
name
	^companyName
! !

!PayrollSystem methodsFor: 'accessing' stamp: ''!
name: aString
	companyName := aString
! !

!PayrollSystem methodsFor: 'actions' stamp: ''!
paycheckString
	| aStream |
	aStream := WriteStream on: (String new: 20).
	self printPaychecksOn: aStream.
	^aStream contents
! !

!PayrollSystem methodsFor: 'actions' stamp: ''!
postPaychecks
	paychecks do: [:each | each post].
	paychecks := #()
! !

!PayrollSystem methodsFor: 'actions' stamp: ''!
printPaychecksOn: aStream 
	paychecks do: [:each | each printOnCheckStream: aStream]
! !

Object subclass: #TaxRule
	instanceVariableNames: 'breakPoints rates'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'CS598rej-Payroll'!

!TaxRule commentStamp: 'm3r 6/6/2008 22:20' prior: 0!
I represent a tax table.  Given a yearly taxable income, I can give the tax rate for it.
I store the tax table as a sequence of amounts, and the tax rate for that amount.
For example, in 1994 the tax rate for a single person (schedule X) was
          $0		15%
  		$22750	28%
  		$55100	31%
		$115000	36%
		$250000	39.6%

breakpoints  <SortedCollection of: integer>
rates	 	<OrderedCollection of: integer>!

!TaxRule class methodsFor: 'instance creation' stamp: ''!
new
	^super new initialize
! !

!TaxRule methodsFor: 'accessing' stamp: ''!
breakPoint: aNumber rate: anotherNumber 
	"Add an entry to the tax table"

	| index |
	index := breakPoints indexOfStartOfIntervalContaining: aNumber.
	breakPoints add: aNumber.
	rates add: anotherNumber beforeIndex: index + 1
! !

!TaxRule methodsFor: 'initialize-release' stamp: ''!
initialize
	breakPoints := SortedCollection new.
	rates := OrderedCollection new.
! !

!TaxRule methodsFor: 'accessing' stamp: ''!
rateFor: aNumber 
	"Return the tax rate for someone whose taxable income is 
	aNumber dollars."

	^rates at: (breakPoints indexOfStartOfIntervalContaining: aNumber)
! !
Employee initialize!
